/***********************************************************************/
/*                                                                     */
/*                               Pio (C)                               */
/*                                                                     */
/*                    Polyglot Input/Output Library                    */
/*                                                                     */
/*               EPI Pomdapi, INRIA Paris - Rocquencourt               */
/*                                                                     */
/*  Copyright 2011-2014 INRIA.                                         */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved. This file is distributed only by permission.  */
/*                                                                     */
/*  Bruno Verlyck <Bruno.Verlyck@inria.fr>                             */
/*                                                                     */
/***********************************************************************/

#include "common.h"
#include "debug_io.h"
#include "pio_ibuff.h"
#include "typed_value.h"
#include "communication.h"
#include "write_communication.h"
#include "read_communication.h"
#include "read_typed_value.h"
#include "write_typed_value.h"
#include "io.h"
#include <unistd.h>

void (*send_communication) (pio_buffer *ob, const COM_t *comm) = write_communication;

/* Return a COM_t to let the caller free the COM */
/* XXX make it a static send_values_internal, with an arg.
   values_allocated to say to use the standard free_communication, when one exists. */
COM_t *send_values (pio_buffer *ob, TV_t **values, size_t nvals) {
  COM_t *com = make_communication (COM_Phrase);
  com->vs.phrase_values = values;
  com->com_nvalues = nvals;
  send_communication (ob, com);
  return com;
}

void send_one_value (pio_buffer *ob, TV_t *value) {
  /* Try to avoid allocating a single pointer */
  TV_t *val_array[1];
  val_array[0] = value;
  TV_t **values = val_array;
  COM_t *com = send_values (ob, values, 1);
  free (com);
}

#define gen_send(native, ob, t) write_typed_value (ob, TV_of_##native (t))
#define send_t(TYPE, container)						\
  void send_##TYPE (pio_buffer *ob, container t) {			\
    gen_send (TYPE, ob, t);						\
  }

send_t (string, struct string *)
send_t (chars, const char *)
send_t (int, int)
send_t (int_couple, struct cint *)
send_t (int_triple, struct tint *)
send_t (float, double)
send_t (float_couple, struct cfloat *)
send_t (float_triple, struct tfloat *)
send_t (string_vector, struct vector)
void send_chars_array (pio_buffer *ob, const char **s, size_t sz) {
  write_typed_value (ob, TV_of_chars_array (s, sz));
}
send_t (int_vector, struct vector)
send_t (int_couple_vector, struct vector)
send_t (int_triple_vector, struct vector)
send_t (float_couple_vector, struct vector)
send_t (float_triple_vector, struct vector)
send_t (float_vector, struct vector)
void send_c_float_matrix (pio_buffer *ob, double *m, size_t szi, size_t szj) {
  write_typed_value (ob, TV_of_c_float_matrix (m, szi, szj));
}
void send_fortran_float_matrix (pio_buffer *ob, double *m, size_t szi, size_t szj) {
  write_typed_value (ob, TV_of_fortran_float_matrix (m, szi, szj));
}
#undef gen_send
#undef send_t

COM_t * (*receive_communication) (pio_input_buffer *ib) = read_communication;

#define fun_name "receive_values"
TV_t **receive_values (pio_input_buffer *ib) {
  COM_t *com = receive_communication (ib);
  switch (com->com_tag) {
  case COM_Phrase:
    return com->vs.phrase_values;
  default:
    failwith (__FILE__ "::" fun_name "communication is not a phrase");
  }
}
#undef fun_name

#define fun_name "receive_one_value"
TV_t *receive_one_value (pio_input_buffer *ib) {
  TV_t **values = receive_values (ib);
  if (values[1] == NULL)
    return values[0];
  failwith (__FILE__ "::" fun_name "communication phrase has more than one value");
}
#undef fun_name

/* Cf. read_communication */
#define gen_receive(native, res_type)					\
  res_type receive_##native (pio_input_buffer *ib) {			\
    jmp_buf rc_buf;							\
    jmp_buf *sv_wrong_communication = wrong_communication_label;	\
    wrong_communication_label = &rc_buf;				\
    jmp_buf *sv_eof_bscanf_label = eof_bscanf_label;			\
    eof_bscanf_label = &rc_buf;						\
    jmp_buf *sv_fail_label = fail_label;				\
    fail_label = &rc_buf;						\
    res_type res;							\
    int come_from = (setjmp (rc_buf));					\
    if (!come_from) {							\
      res = to_##native (read_typed_value (ib));			\
    } else {								\
      const char *exc_msg =						\
	(come_from == (int)((long)(&wrong_communication_label)))	\
	? wrong_communication_msg					\
	: (come_from == (int)((long)(&eof_bscanf_label)))		\
	? "Eof in ibuf_scanf"						\
	: (come_from == (int)((long)(&fail_label)))			\
	? failure_msg							\
	: (assert (0), "unknown exception in gen_receive");		\
      snprintf (failbuf, sizeof (failbuf), __FILE__ ":: spurious exception %s\n", exc_msg); \
      pio_fatal_error (failbuf);					\
    }									\
    wrong_communication_label = sv_wrong_communication;			\
    eof_bscanf_label = sv_eof_bscanf_label;				\
    fail_label = sv_fail_label;						\
    return res;								\
}

gen_receive (string, struct string)
#define funname "receive_chars"
const char *receive_chars(pio_input_buffer *ib) {
  struct string s = receive_string(ib);
  if (s.ssize != strlen (s.str))
    failwithf (__FILE__ "::" funname ": string starting with `%s' contains NUL(s)", s.str);
  return s.str;
}
#undef funname
gen_receive (int, int)
gen_receive (int_couple, struct cint)
gen_receive (int_triple, struct tint)
gen_receive (float, double)
gen_receive (float_couple, struct cfloat)
gen_receive (float_triple, struct tfloat)
gen_receive (string_vector, struct string *)
gen_receive (int_vector, int *)
gen_receive (int_couple_vector, struct cint *)
gen_receive (int_triple_vector, struct tint *)
gen_receive (float_vector, double *)
gen_receive (float_couple_vector, struct cfloat *)
gen_receive (float_triple_vector, struct tfloat *)
gen_receive (c_float_matrix, double *)
gen_receive (fortran_float_matrix, double *)
